@@ -3,7 +3,7 @@ from __future__ import absolute_import
 import functools
 import operator
 
-from .bindings import ffi, C, GSS_ERROR, _buf_to_str
+from .bindings import ffi, C, GSS_ERROR, GSS_SUPPLEMENTARY_INFO, _buf_to_str
 from .error import GSSException, _exception_for_status
 from .names import MechName, Name
 from .oids import OID
@@ -19,6 +19,17 @@ def _release_gss_ctx_id_t(context):
         )
 
 
+def _status_bits(retval):
+    supplementary_info = GSS_SUPPLEMENTARY_INFO(retval)
+    return tuple(
+        f for f in (
+            C.GSS_S_DUPLICATE_TOKEN, C.GSS_S_OLD_TOKEN, C.GSS_S_UNSEQ_TOKEN, C.GSS_S_GAP_TOKEN
+        )
+        if f & supplementary_info
+    )
+
+
+
 class Context(object):
     """
     Represents a GSSAPI security context. This class manages establishing the context with a peer,
@@ -26,8 +37,8 @@ class Context(object):
     messages exchanged with the peer.
 
     This class is not particularly useful in itself, but holds common functionality for all types
-    of context. To use a context as the initiator or acceptor, :class:`InitContext` or
-    :class:`AcceptContext` should be used, respectively.
+    of context. To use a context as the initiator or acceptor, create an :class:`InitContext` or
+    :class:`AcceptContext`, respectively.
 
     .. py:attribute:: established
 
@@ -149,6 +160,8 @@ class Context(object):
 
         :param message: The message to calculate a MIC for
         :type message: bytes
+        :param qop_req: The quality of protection required. It is recommended to not change this
+            from the default as most GSSAPI implementations do not support it.
         :returns: A MIC for the message calculated using this security context's cryptographic keys
         :rtype: bytes
         """
@@ -183,19 +196,30 @@ class Context(object):
             if output_token_buffer[0].length != 0:
                 C.gss_release_buffer(minor_status, output_token_buffer)
 
-    def verify_mic(self, message, mic):
+    def verify_mic(self, message, mic, supplementary=False):
         """
         Takes a message integrity code (MIC) that has been generated by the peer application for a
         given message, and verifies it against a message, using this security context's
         cryptographic keys.
 
+        If the `supplementary` parameter is False (default), returns the quality of protection
+        only (the ``qop_state`` value). If `supplementary` is True, returns a tuple ``(qop_state,
+        supplementary_info)`` where ``supplementary_info`` is a tuple containing zero or more of
+        the constants :const:`~gssapi.S_DUPLICATE_TOKEN`, :const:`~gssapi.S_OLD_TOKEN`,
+        :const:`~gssapi.S_UNSEQ_TOKEN` and :const:`~gssapi.S_GAP_TOKEN`. The supplementary info
+        tells the caller whether a replayed or out-of-sequence message was detected.
+
         :param message: The message the MIC was calculated for
         :type message: bytes
         :param mic: The MIC calculated by the peer
         :type mic: bytes
-        :returns: the quality of protection (qop_state)
-        :raises: GSSException if :attr:`integrity_negotiated` is false, or if the verification
-            fails indicating the message was modified
+        :param supplementary: Whether to also return supplementary info.
+        :type supplementary: bool
+        :returns: ``qop_state`` if `supplementary` is False, or ``(qop_state,
+            supplementary_info)`` if `supplementary` is True.
+        :raises: :exc:`~gssapi.error.GSSException` if :attr:`integrity_negotiated` is false, or
+            :exc:`~gssapi.GSSCException` if the verification fails indicating the message was
+            modified.
         """
         if not (self.flags & C.GSS_C_INTEG_FLAG):
             raise GSSException("No integrity protection negotiated.")
@@ -225,7 +249,10 @@ class Context(object):
                 raise _exception_for_status(retval, minor_status[0], self.mech_type)
             else:
                 raise _exception_for_status(retval, minor_status[0])
-        return qop_state[0]
+        if supplementary:
+            return qop_state[0], _status_bits(retval)
+        else:
+            return qop_state[0]
 
     def wrap(self, message, conf_req=True, qop_req=C.GSS_C_QOP_DEFAULT):
         """
@@ -237,6 +264,8 @@ class Context(object):
         :type message: bytes
         :param conf_req: Whether to require confidentiality (encryption)
         :type conf_req: bool
+        :param qop_req: The quality of protection required. It is recommended to not change this
+            from the default as most GSSAPI implementations do not support it.
         :returns: the wrapped message in a token suitable for passing to :meth:`unwrap`
         :rtype: bytes
         :raises: GSSException if integrity protection is not available
@@ -283,20 +312,33 @@ class Context(object):
             if output_token_buffer[0].length != 0:
                 C.gss_release_buffer(minor_status, output_token_buffer)
 
-    def unwrap(self, message, conf_req=True, qop_req=None):
+    def unwrap(self, message, conf_req=True, qop_req=None, supplementary=False):
         """
         Takes a token that has been generated by the peer application with :meth:`wrap`, verifies
         and optionally decrypts it, using this security context's cryptographic keys.
 
+        If the `supplementary` parameter is False (default), returns the unwrapped message only.
+        If `supplementary` is True, returns a tuple ``(unwrapped_message, supplementary_info)``
+        where ``supplementary_info`` is a tuple containing zero or more of the constants
+        :const:`~gssapi.S_DUPLICATE_TOKEN`, :const:`~gssapi.S_OLD_TOKEN`,
+        :const:`~gssapi.S_UNSEQ_TOKEN` and :const:`~gssapi.S_GAP_TOKEN`. The supplementary info
+        tells the caller whether a replayed or out-of-sequence message was detected.
+
         :param message: The wrapped message token
         :type message: bytes
         :param conf_req: Whether to require confidentiality (encryption)
         :type conf_req: bool
-        :returns: the verified and decrypted message
-        :raises: GSSException if :attr:`integrity_negotiated` is false, or if the verification or
-            decryption fails, if the message was modified, or if confidentiality was required
-            (`conf_req` was True) but the message did not have confidentiality protection applied
-            (was not encrypted).
+        :param qop_req: The quality of protection required. It is recommended to not change this
+            from the default None as most GSSAPI implementations do not support it.
+        :param supplementary: Whether to also return supplementary info.
+        :type supplementary: bool
+        :returns: the verified and decrypted message if `supplementary` is False, or a tuple
+            ``(unwrapped_message, supplementary_info)`` if `supplementary` is True.
+        :raises: :exc:`~gssapi.error.GSSException` if :attr:`integrity_negotiated` is false, or if
+            the verification or decryption fails, if the message was modified, or if
+            confidentiality was required (`conf_req` was True) but the message did not have
+            confidentiality protection applied (was not encrypted), or if the `qop_req`
+            parameter was set and it did not match the QOP applied to the message.
         """
         if not (self.flags & C.GSS_C_INTEG_FLAG):
             raise GSSException("No integrity protection negotiated.")
@@ -332,7 +374,10 @@ class Context(object):
                 raise GSSException("No confidentiality protection.")
             if qop_req is not None and qop_req != qop_state[0]:
                 raise GSSException("QOP {0} does not match required value {1}.".format(qop_state[0], qop_req))
-            return output
+            if supplementary:
+                return output, _status_bits(retval)
+            else:
+                return output
         finally:
             if output_buffer[0].length != 0:
                 C.gss_release_buffer(minor_status, output_buffer)
@@ -370,6 +415,41 @@ class Context(object):
 
         return max_input_size[0]
 
+    def process_context_token(self, context_token):
+        """
+        Provides a way to pass an asynchronous token to the security context, outside of the normal
+        context-establishment token passing flow. This method is not normally used, but some
+        example uses are:
+
+        * when the initiator's context is established successfully but the acceptor's context isn't
+          and the acceptor needs to signal to the initiator that the context shouldn't be used.
+        * if :meth:`delete` on one peer's context returns a final token that can be passed to the
+          other peer to indicate the other peer's context should be torn down as well (though it's
+          recommended that :meth:`delete` should return nothing, i.e. this method should not be
+          used by GSSAPI mechanisms).
+
+        :param context_token: The context token to pass to the security context
+        :type context_token: bytes
+        :raises: :exc:`~gssapi.error.DefectiveToken` if consistency checks on the token failed.
+            :exc:`~gssapi.error.NoContext` if this context is invalid.
+            :exc:`~gssapi.error.GSSException` for any other GSSAPI errors.
+        """
+        minor_status = ffi.new('OM_uint32[1]')
+        context_token_buffer = ffi.new('gss_buffer_desc[1]')
+        context_token_buffer[0].length = len(context_token)
+        c_str_context_token = ffi.new('char[]', context_token)
+        context_token_buffer[0].value = c_str_context_token
+        retval = C.gss_process_context_token(
+            minor_status,
+            self._ctx[0],
+            context_token_buffer
+        )
+        if GSS_ERROR(retval):
+            if minor_status[0] and self.mech_type:
+                raise _exception_for_status(retval, minor_status[0], self.mech_type)
+            else:
+                raise _exception_for_status(retval, minor_status[0])
+
     def export(self):
         """
         This method deactivates the security context for the calling process and returns an
@@ -512,13 +592,16 @@ class Context(object):
         """
         Delete a security context. This method will delete the local data structures associated
         with the specified security context, and may return an output token, which when passed to
-        :meth:`process_context_token` on the peer will instruct it to do likewise.
+        :meth:`process_context_token` on the peer may instruct it to also delete its context.
+
+        RFC 2744 recommends that GSSAPI mechanisms do not emit any output token when they're
+        deleted, so this behaviour could be considered deprecated.
 
         After this method is called, this security context will become invalid and should not be
         used in any way.
 
-        :returns: An output token if one should be passed to :meth:`process_context_token` on the
-            peer, otherwise an empty bytestring.
+        :returns: An output token if one was emitted by the GSSAPI mechanism, otherwise an empty
+            bytestring.
         :rtype: bytes
         """
 
@@ -618,7 +701,7 @@ class InitContext(Context):
         :type input_token: bytes
         :returns: either a byte string with the next token to send to the acceptor,
             or None if there is no further token to send to the acceptor.
-        :raises: GSSException
+        :raises: :exc:`~gssapi.error.GSSException` if there is an error establishing the context.
         """
 
         minor_status = ffi.new('OM_uint32[1]')
@@ -700,9 +783,9 @@ class AcceptContext(Context):
     context on the acceptor side. The initiator is normally authenticated as part of the context
     establishment process, though some mechanisms support anonymous peers.
 
-    :param cred: The credential to use for the acceptor. Omit this parameter or pass
-        :const:`gssapi.C_NO_CREDENTIAL` to use the default acceptor credentials (e.g. any principal
-        in the default keytab, when the Kerberos mechanism is used).
+    :param cred: The credential to use for the acceptor. Omit this parameter to use the default
+        acceptor credentials (e.g. any principal in the default keytab, when the Kerberos mechanism
+        is used).
     :type cred: :class:`~gssapi.creds.Credential`
     :param input_chan_bindings: Optional channel bindings object, to bind this security context to
         an underlying communications channel.
@@ -757,7 +840,7 @@ class AcceptContext(Context):
         :type input_token: bytes
         :returns: either a byte string with the next token to send to the initiator,
             or None if there is no further token to send to the initiator.
-        :raises: GSSException
+        :raises: :exc:`~gssapi.error.GSSException` if there is an error establishing the context.
         """
         minor_status = ffi.new('OM_uint32[1]')
         input_token_buffer = ffi.new('gss_buffer_desc[1]')
